#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from .common import *
from . import implem

################################################################################
# Dummy init
################################################################################

@testcase(SixLoWPAN_ND)
def init( link1 ):
	ereq = pack (
		IPv6 (src = "::", dst = "::" ),
		ICMPv6EchoRequest(),
		"init()"
	)
	link1.send(ereq)

################################################################################
# Test_LPND_1_1_1a
###############################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_1a(NUT , TIMER1 , LBR1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router solicitation (RS) " )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) / 
		ICMPv6RSol( )
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1_1_1b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_1b(NUT , TIMER1 , LBR1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq = pack (
		IPv6 (src = LBR1.lladdr, dst = IPV6_ALL_NODES),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.1b"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "LBR-1 sends a ICMPv6 Echo Request to the  IPV6_ALL_NODES (", IPV6_ALL_NODES,") address")
	link1.send(ereq)

	step (3, "We expect an ICMPv6 Echo Reply")
	t.start(TIMER1)	

	link1.receive (IPv6()/ICMPv6EchoReply())
	set_verdict ("pass")

################################################################################
# Test_LPND_1_1_2a
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_2a(NUT , TIMER1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect an ICMPv6 Router solicitation")
	t.start(TIMER1)
	with alt:	#FIXME: removable
		@link1.receive (
			IPv6( dst = IPV6_ALL_ROUTERS ) /
			ICMPv6RSol( opt = superop ([
				ICMPv6SLL( hw=NUT.hwaddr ) 
				] )
			)
		)
		def _(value):
			if not value[IPv6] in IPv6(src = "::"):
				set_verdict ("pass")
			else:
				set_verdict ("fail")


################################################################################
# Test_LPND_1_1_2b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_2b(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr, hl = 255 ) ,
		ICMPv6RAdv( rlt = 30 , hp = 255 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr) , 
			SixPI( pf=PRFX1 , l = 0 ) , 
			]
		)
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect an ICMPv6 Router solicitation to send an ICMPv6 Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) 
			] )
		)
	)
	link1.send(ra)

	step (3, "We expect an ICMPv6 Neighboor solicitation and another ICMPv6 Router solicitation")
	t.start(30)
	link1.receive (
		IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			SixARO( )
			] )
		)
	)
	set_verdict ("pass")

	link1.receive (
		IPv6( src = NUT.lladdr ) /
		ICMPv6RSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) 
			] )
		)
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1_1_2c
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_2c(NUT , TIMER1 , LBR1 , PRFX1, link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt =  [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI(pf=PRFX1,vlt = 30 , plt = 30 )

			]
		)
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect an ICMPv6 Router solicitation to send an ICMPv6 Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) 
			] )
		)
	)
	link1.send(ra)

	step (3, "We expect an ICMPv6 Neighboor solicitation and another ICMPv6 Router solicitation")
	t.start(30)
	
	link1.receive (
		IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			SixARO()
			] )
		)
	)
	set_verdict ("pass")

	link1.receive (
		IPv6( src = NUT.lladdr , dst = LBR1.lladdr ) /
		ICMPv6RSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr )
			] )
		)
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1_1_2d (TODO/FIXME)
################################################################################
#TODO: verify timer value
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_2d(NUT , TIMER1 , TIMER2, LBR1 , PRFX1, link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect six (6) ICMPv6 Router solicitations")
	t.start(180) # 2x10 + 20 + 40 + max (60 , 80 ) + margin: 20 seconds
	timers = []

	for x in range(0,6):
		with alt: #FIXME: removable
			@link1.receive (
				IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
				ICMPv6RSol( opt = superop ( [
					ICMPv6SLL( hw=NUT.hwaddr ) 
					] )
				)
			)
			def _(value):
				timers.append( t._Timer__clock.time())

	margin = 1

	for x in range(1,6): # we ignore the day of the first RS
		delay = timers[x] - timers[x-1]
		step(2,"RS #"+repr(x+1)+" received after "+repr(delay) + " seconds after RS#"+repr(x))

		if x == 1 or x == 2:
			if delay >= 10 - margin :
				set_verdict("pass")
			else:
				set_verdict("fail")				
		elif x == 3:
			if delay >= 20 - margin:
				set_verdict("pass")
			else:
				set_verdict("fail")
		elif x == 4:
			if delay >= 40 - margin:
				set_verdict("pass")
			else:
				set_verdict("fail")
		elif x == 5:
			if delay >= 60 - margin: # 2x40 = 80. "However, it is useful to have a maximum retransmission timer of 60 seconds"
				set_verdict("pass")
			else:
				set_verdict("fail")	

	step (3, "We send an ICMPv6 Router Advertisement and don't except any Router Solicitation")
	link1.send(ra)
	t2.start(TIMER2)
	with alt:
		@link1.receive (
			IPv6( src = NUT.lladdr ) /
			ICMPv6RSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) 
				] )
			)
		)
		def _():
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1_1_3a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_3a(NUT , TIMER1 , LBR1 , link1, action, PRFX1):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 , l=0 )
			]
		)
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send an ICMPv6 Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)
	t.stop()

	step (3, "We expect a Neighbour Solicitation with an Address Registration Option")
	t.start(TIMER1)

	link1.receive (
		IPv6( src = NUT.addr , dst=LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			SixARO( eui = NUT.eui64 )
			] )
		)
	)
	set_verdict ("pass")


################################################################################
# Test_LPND_1_1_3b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_3b(NUT , TIMER1 , LBR1 , link1, action, PRFX1,TIMER2):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 , l=1 )
			]
		)
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send an ICMPv6 Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)
	t.stop()

	step (3, "We expect that the NUT don't send any Neighbour Solicitation")
	t2.start(TIMER2)
	with alt:
		@link1.receive (
			IPv6( src = NUT.addr , dst=LBR1.lladdr ) /
			ICMPv6NSol( )
		)
		def _():
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1_1_3c
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_3c(NUT , TIMER1 , LBR1 , link1, action, PRFX1, PRFX2):


	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 ) ,
			SixPI( pf=PRFX2 ) ,
			]
		)
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send an ICMPv6 Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)
	t.stop()

	step (3, "We expect two Neighbour Solicitation with an Address Registration Option to sends two Neighbour Advertisement")
	t.start(TIMER1)

	addr1 = IPv6Prefix(PRFX1).make_address(Eui64Address(NUT.eui64))
	addr2 = IPv6Prefix(PRFX2).make_address(Eui64Address(NUT.eui64))

	received_neighboursolicitions = receive_unordered (
		link1 , 
		IPv6( src =  addr1 , dst=LBR1.lladdr ) / ICMPv6NSol( opt = superop ( [ SixARO( eui = NUT.eui64	) ] ) ) ,
		IPv6( src =  addr2 , dst=LBR1.lladdr ) / ICMPv6NSol( opt = superop ( [ SixARO( eui = NUT.eui64	) ] ) ) ,
		)
	set_verdict ("pass")
	for ns in received_neighboursolicitions:
		na = pack (
			IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
			ICMPv6NAdv( opt = [
				SixARO( lt = ns[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 0 , 
					eui= ns[IPv6][ICMPv6]["opt"][SixARO]["eui"])
				]
			)
		)
		link1.send(na)

	step (4, "We sends two ICMPv6 Echo Request and Expect two ICMPv6 Echo Reply")
	t.start(TIMER1)
	for addr in [ addr1 , addr2 ]:
		ereq = pack (
			IPv6 (src = LBR1.addr, dst = addr),
			ICMPv6EchoRequest(),
			"6lowpan-nd-1.1.3c-"+str(addr)
		)
		link1.send(ereq)

	receive_unordered (
		link1 , 
		IPv6( src =  addr1 , dst=LBR1.addr ) / ICMPv6EchoReply(),
		IPv6( src =  addr2 , dst=LBR1.addr ) / ICMPv6EchoReply(),
		)

################################################################################
# Test_LPND_1_1_4a
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_4a(NUT , TIMER1 , TIMER2, LBR1 , link1, action, PRFX1):
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 ,  m=0,  opt = [ 
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send an ICMPv6 Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighbour solicitation")
	t.start(TIMER1)
	link1.receive (
		IPv6( src=NUT.addr , dst=LBR1.lladdr) /
		ICMPv6NSol( opt = superop ( [
			SixARO( eui = NUT.eui64 ) ,
			ICMPv6SLL( hw=NUT.hwaddr )
			] )
		)
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1_1_5a
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_5a(NUT , TIMER1, TIMER2 , LBR1 , PRFX1 , link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )


	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5a"
	)

	step (1, "We perform the common test setup 1.1 with a router lifetime set to 30" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 , routerlt = 30 )

	step (2, "We expect a Router Solicitation Before 30 seconds" )
	t.start(30)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = LBR1.lladdr ) /
		ICMPv6RSol( )
	)
	set_verdict ("pass")
	t.stop()

	step (3, "We wait for 30 seconds..." )
	t.start (30)
	t.timeout()

	step (4, "We sends an Echo Request and don't expect any answer" )
	t2.start(TIMER2)
	link1.send(ereq)

	with alt:
		@link1.receive (
			IPv6( src=NUT.addr ) /
			ICMPv6ERep()
		)
		def _():
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1_1_5b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_5b(NUT , TIMER1 , TIMER2 , LBR1 , PRFX1, link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5b"
	)

	step (1, "We perform the common test setup 1.1 with a prefix lifetime set to 30" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 , prefixlt = 30 )

	step (2, "We expect a Router Solicitation Before 30 seconds" )
	t.start(30)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = LBR1.lladdr ) /
		ICMPv6RSol( )
	)
	set_verdict ("pass")
	t.stop()

	step (3, "We wait for 30 seconds..." )
	t.start (30)
	t.timeout()

	step (4, "We sends an Echo Request and don't expect any answer" )
	link1.send(ereq)
	t2.start(TIMER2)
	with alt:
		@link1.receive (
			IPv6( src=NUT.addr ) /
			ICMPv6ERep()
		)
		def _():
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1_1_5c
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_5c(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq1 = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5c-1"
	)
	ereq2 = pack (
		SixLowpanIPHC() ,
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5c-2"
	)
	ereq3 = pack (
		SixLowpanIPHC() ,
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5c-3"
	)


	step (1, "We perform the common test setup 1.1b with a context lifetime set to 30" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1 , contextlt = 30 )

	step (2, "We expect a Router Solicitation Before 30 seconds" )
	t.start(30)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = LBR1.lladdr ) /
		ICMPv6RSol( )
	)
	t.stop()

	step (3, "We wait for 30 seconds..." )
	t.start (30)
	t.timeout()

	step (4, "We sends two ICMPv6 Echo Request (One compressed, one not compressed). We expect two uncompressed ICMPv6 Echo Reply" )
	link1.send(ereq1)
	link1.send(ereq2)
	t.start(TIMER1)

	receive_unordered ( link1 ,
		(
			SixLowpanIPv6() /
			IPv6( src=NUT.addr ) /
			ICMPv6ERep()
		),
		(
			SixLowpanIPHC() /
			IPv6( src=NUT.addr ) /
			ICMPv6ERep()
		)
	)

	step (5, "We wait for 60 seconds..." )
	t.start(60)
	t.timeout()

	step (6, "We sends a compressed ICMPv6 Echo Request. We expect a uncompressed ICMPv6 Echo Reply" )
	link1.send(ereq3)
	t.start(TIMER1)
	link1.receive (
		SixLowpanIPv6() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)
	set_verdict("pass")


################################################################################
# Test_LPND_1_1_5d
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_5d(NUT , TIMER1 , LBR1 , PRFX1 , SLEEP1, link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 , vlt = 60 ) ,
			SixCO( c=0 , clen=64, pf=PRFX1 , lt=60)
			]
		)
	)

	ereq1 = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5d-1"
	)
	ereq2 = pack (
		SixLowpanIPHC() ,
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5d-2"
	)
	ereq3 = pack (
		SixLowpanIPHC() ,
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.5d-3"
	)

	step (1, "We perform the common test setup 1.1b with a context lifetime set to 30" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1 , contextlt = 30 )

	step (2, "We sends a Router Adverstisement with the Compress (C) bit set to 0" )
	link1.send(ra)

	step (3, "We sends two ICMPv6 Echo Request (One compressed, one not compressed). We expect two uncompressed ICMPv6 Echo Reply" )
	link1.send(ereq1)
	link1.send(ereq2)
	t.start(TIMER1)

	receive_unordered ( link1 ,
		(
			SixLowpanIPv6() /
			IPv6( src=NUT.addr ) /
			ICMPv6ERep()
		),
		(
			SixLowpanIPv6() /
			IPv6( src=NUT.addr ) /
			ICMPv6ERep()
		)
	)

	step (4, "We wait for 60 seconds..." )
	t.start (60)
	t.timeout()

	step (5, "We sends a compressed ICMPv6 Echo Request. We expect a uncompressed ICMPv6 Echo Reply" )
	link1.send(ereq3)
	t.start(TIMER1)
	link1.receive (
		SixLowpanIPv6() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)
	set_verdict("pass")


################################################################################
# Test_LPND_1_1_5e
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_5e(NUT , TIMER1 , PRFX1 , LBR1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 , routerlt = 30 )
	
	step (2, "We wait for 30 seconds")
	t.start(30)
	t.stop()

	step (2, "We expect an unicast router solicitation and a multicast one")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = LBR1.lladdr ) /
		ICMPv6RSol( )
	)
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol()
		)

	set_verdict ("pass")


################################################################################
# Test_LPND_1_1_6a
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_6a(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ra2 = pack (
		IPv6( src = LR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)
	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send two (2) Router advertisements" )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)

	link1.send(ra)
	link1.send(ra2)

	step (3, "We expect two (2) Neighbor solicitation")
	t.start(TIMER1)
	receive_unordered ( link1 ,
		(
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( eui = NUT.eui64 )
				] )
			)
		),
		(
			IPv6( src = NUT.addr , dst = LR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( eui = NUT.eui64  )
				] )
			)
		) 
	)
	set_verdict ("pass")	


################################################################################
# Test_LPND_1_1_6b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_6b(NUT , TIMER1 , LBR1 , PRFX1, SLEEP1, link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 , vlt = 60 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.6b"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighboor Solicitation to send a Neighboor Advertisement")
	t.start(TIMER1)
	with alt:
		@link1.receive (
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol(opt = superop ( [
				SixARO( )
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv( opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 0,
						eui= value[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)
			link1.send(na)

	step (3, "We sends an ICMPv6 Echo Request and expect for an Answer")
	link1.send(ereq)
	t.start(TIMER1)
	link1.receive ( IPv6()/ICMPv6EchoReply() )
	set_verdict ("pass")


################################################################################
# Test_LPND_1_1_6c
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_6c(NUT , TIMER1 , LBR1 , link1, action, PRFX1):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt =  [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 , vlt = 60 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.6c"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighboor Solicitation to send a Neighboor Advertisement")
	t.start(TIMER1)
	with alt:
		@link1.receive (
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv( opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 1 , 
						eui= value[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)
			link1.send(na)

	step (3, "We sends an ICMPv6 Echo Request and don't expect for an answer")
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1_1_6d
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_6d(NUT , TIMER1 , LBR1 , LR1, link1, action, PRFX1):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ra2 = pack (
		IPv6( src = LR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv(opt = [
			ICMPv6SLL( hw=LR1.hwaddr ) ,
			SixPI( pf=PRFX1  )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.6d"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send two (2) Router advertisements" )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)
	link1.send(ra2)

	step (3, "We expect two (2) Neighboor Solicitations to send two (2) Neighboor Advertisements " )
	received_neighboursolicitions = receive_unordered ( link1 ,
		( 
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol(opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		),
		( 
			IPv6( src = NUT.addr , dst = LR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		)
	)
	ns = received_neighboursolicitions.pop()
	ns2 = received_neighboursolicitions.pop()

	na = pack (
		IPv6( src = ns[IPv6]["dst"] , dst = NUT.lladdr ) ,
		ICMPv6NAdv(opt = [
			SixARO( lt = ns[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 0,
				eui= ns[IPv6][ICMPv6]["opt"][SixARO]["eui"])
			]
		)
	)

	na2 = pack (
		IPv6( src = ns2[IPv6]["dst"] , dst = NUT.lladdr ) ,
		ICMPv6NAdv(opt = [
			SixARO( lt = ns2[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 1,
					eui= ns2[IPv6][ICMPv6]["opt"][SixARO]["eui"])
			]
		)
	)

	link1.send(na)
	link1.send(na2)

	step (4, "We expect a Neighboor Solicitations" )
	t.start(TIMER1)
	link1.receive ( 
		IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			SixARO( lt = 0 )
			] )
		)
	)

	step (5, "We sends an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1_1_6e
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_6e(NUT , TIMER1 , LBR1 , link1, action, PRFX1):
	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( rlt = 60 , opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 , vlt = 60 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.6e"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighboor Solicitation to send a Neighboor Advertisement")
	t.start(TIMER1)
	with alt:	#FIXME: removable
		@link1.receive (
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv( opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 2 ,
						eui= value[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)
			link1.send(na)

	step (4, "We expect a multicast Router Solicitation")
	t.start(TIMER1)
	link1.receive ( 
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			] )
		)
	)

	step (5, "We sends an ICMPv6 Echo Request and don't expect any answer")
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)

	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1_1_6f
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_6f(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ra2 = pack (
		IPv6( src = LR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv(opt = [
			ICMPv6SLL( hw=LR1.hwaddr ) ,
			SixPI( pf=PRFX1  )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.6f"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send two (2) Router advertisements" )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)
	link1.send(ra2)
		
	step (3, "We expect two (2) Neighboor Solicitations to send two (2) Neighboor Advertisements " )
	received_neighboursolicitions = receive_unordered ( link1 ,
		( 
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol(opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		),
		( 
			IPv6( src = NUT.addr , dst = LR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		)	
	)

	ns = received_neighboursolicitions.pop()
	ns2 = received_neighboursolicitions.pop()

	na = pack (
		IPv6( src = ns[IPv6]["dst"] , dst = NUT.lladdr ) ,
		ICMPv6NAdv(opt = [
			SixARO( lt = ns[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 0 ,
				eui= ns[IPv6][ICMPv6]["opt"][SixARO]["eui"])
			]
		)
	)

	na2 = pack (
		IPv6( src = ns2[IPv6]["dst"] , dst = NUT.lladdr ) ,
		ICMPv6NAdv(opt = [
			SixARO( lt = ns2[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 2 ,
				eui= ns2[IPv6][ICMPv6]["opt"][SixARO]["eui"])
			]
		)
	)

	link1.send(na)
	link1.send(na2)

	step (4, "We sends an ICMPv6 Echo Request and expect an answer via LR-1")
	link1.send(ereq)
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.addr , dst = LBR1.addr )	/ ICMPv6EchoReply() , dst = LR1.hwaddr )
	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.6g
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_6g(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 )

	step (2, "NUT action")
	action ("Press enter, then shutdown the NUT")

	step (3, "We except a Neighboor Solicitation" )
	t.start(TIMER1)
	link1.receive ( 
		IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			SixARO( lt = 0 )
			] )
		)
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.7a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_7a(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):
	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.7a"
	)
	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighboor Solicitation to send a Neighboor Advertisement")
	t.start(TIMER1)
	with alt:	#FIXME: removable
		@link1.receive (
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				SixARO()
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv( opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 0 , len = 1,
						eui= value[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)
			link1.send(na)

	step (4, "We sends an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1.1.7b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_7b(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):
	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.7b"
	)
	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighboor Solicitation to send a Neighboor Advertisement")
	t.start(TIMER1)
	with alt:	#FIXME: removable
		@link1.receive (
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO()
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv( opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 0 , eui = '0123456789012345' )
					]
				)
			)
			link1.send(na)

	step (4, "We sends an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1.1.7c
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_7c(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1)
			]
		)
	)



	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.7c"
	)
	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighboor Solicitation to send a Neighboor Advertisement")
	t.start(TIMER1)
	with alt:	#FIXME: removable
		@link1.receive (
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO()
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv( opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"], st = 0 ,
						eui = value[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)
			link1.send(na)

	step (4, "We sends an ICMPv6 Echo Request and expect an answer" )
	link1.send(ereq)
	t.start(TIMER1)

	link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.8a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_8a(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.8a"
	)
	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (3, "We expect a Neighboor Solicitation and don't send any Neighboor Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			SixARO()
			] )
		)
	)

	step (4, "We sends an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1.1.8b
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_8b(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.8b"
	)
	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router Advertisement")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)

	step (2, "We expect a Neighboor Solicitation to send a Neighboor Advertisement")
	t.start(TIMER1)
	with alt:	#FIXME: removable
		@link1.receive (
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO()
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv( opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 80 ,
						eui= value[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)
			link1.send(na)

	step (5, "We sends an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1.1.8c 
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_8c(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ra2 = pack (
		IPv6( src = LR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv(opt = [
			ICMPv6SLL( hw=LR1.hwaddr ) ,
			SixPI( pf=PRFX1  )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.8c"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send two (2) Router advertisements" )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)
	link1.send(ra2)
		
	step (3, "We expect two (2) Neighboor Solicitations to send two (2) Neighboor Advertisements" )
	received_neighboursolicitions = receive_unordered( link1 ,
		( 
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol(opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		),
		( 
			IPv6( src = NUT.addr , dst = LR1.lladdr ) /
			ICMPv6NSol( opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		)
	)


	ns = received_neighboursolicitions.pop()
	ns2 = received_neighboursolicitions.pop()

	if ns[IPv6]["dst"] == IPv6Address(LBR1.lladdr):
		lt = ns[IPv6][ICMPv6]["opt"][SixARO]["lt"]
		l2 = ns2[IPv6][ICMPv6]["opt"][SixARO]["lt"]
	elif ns2[IPv6]["dst"] == IPv6Address(LBR1.lladdr):
		lt = ns2[IPv6][ICMPv6]["opt"][SixARO]["lt"]
		lt2 = ns[IPv6][ICMPv6]["opt"][SixARO]["lt"]	
	else:
		set_verdict("Error")

	na = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6NAdv(opt = [
			SixARO( lt = lt , st = 0 , eui = ns[IPv6][ICMPv6]["opt"][SixARO]["eui"])
			]
		)
	)

	na2 = pack (
		IPv6( src = LR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6NAdv(opt = [
			SixARO( lt = lt2 , st = 1 , eui = ns[IPv6][ICMPv6]["opt"][SixARO]["eui"])
			]
		)
	)

	link1.send(na)
	link1.send(na2)
	step (4, "We expect a Neighboor Solicitations" )
	t.start(TIMER1)
	link1.receive ( 
		IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			SixARO( lt = 0 )
			] )
		)
	)

	step (5, "We sends an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")

################################################################################
# Test_LPND_1.1.8d
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_8d(NUT , TIMER1 , LBR1 , LR1 , PRFX1 , link1, action):

	t = Timer()
	t2 = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 )
			]
		)
	)

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.8d"
	)

	step (1, "Preamble")
	action ("Initialize the network of the NUT")

	step (2, "We expect a Router Solicitation to send a Router advertisement" )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)
	link1.send(ra)
			
	step (3, "We expect a Neighboor Solicitations to send a Neighboor Advertisement" )
	with alt:	#FIXME: removable
		@link1.receive ( 
			IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
			ICMPv6NSol(opt = superop ( [
				ICMPv6SLL( hw=NUT.hwaddr ) ,
				SixARO( )
				] )
			)
		)
		def _(value):
			na = pack (
				IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
				ICMPv6NAdv(opt = [
					SixARO( lt = value[IPv6][ICMPv6]["opt"][SixARO]["lt"] , st = 2 ,
						eui= value[IPv6][ICMPv6]["opt"][SixARO]["eui"])
					]
				)
			)
			link1.send(na)

	step (4, "We expect a Router Solicitation" )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)

	step (5, "We sends an ICMPv6 Echo Request and don't expect any answer" )
	link1.send(ereq)
	t.stop()
	t2.start(TIMER1)
	with alt:
		@link1.receive ( IPv6( src=NUT.addr , dst = LBR1.addr )/ICMPv6EchoReply() )
		def _( ):
			set_verdict ("fail")

		@t2.timeout ()
		def _():
			set_verdict ("pass")


################################################################################
# Test_LPND_1.1.9a
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_9a(NUT , TIMER1 , LBR1 , LR1, HOST2 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 )

	ereq1 = pack (
		IPv6 (src = LBR1.lladdr, dst = NUT.lladdr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.9a-LBR1"
	)
	ereq2 = pack (
		IPv6 (src = LR1.lladdr, dst = NUT.lladdr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.9a-LR1"
	)
	ereq3 = pack (
		IPv6 (src = HOST2.lladdr, dst = NUT.lladdr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.9a-HOST2"
	)

	step (2, "We sends an ICMPv6 Echo Request" )
	link1.send(ereq1)
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.lladdr , dst = LBR1.lladdr ) / ICMPv6EchoReply() , dst=LBR1.hwaddr )
	set_verdict ("pass")

	step (3, "We sends an ICMPv6 Echo Request" )
	link1.send(ereq2)
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.lladdr , dst = LR1.lladdr ) / ICMPv6EchoReply() , dst=LR1.hwaddr )
	set_verdict ("pass")

	step (4, "We sends an ICMPv6 Echo Request" )
	link1.send(ereq3)
	t.start(TIMER1)
	link1.receive ( IPv6( src=NUT.lladdr , dst = HOST2.lladdr ) / ICMPv6EchoReply() , dst=HOST2.hwaddr )

	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.9b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_9b(NUT , TIMER1 , LBR1 , PRFX1 , PRFX2 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 )

	ereq = pack (
		IPv6 (src = IPv6Address(PRFX2.replace("::","") + "::1"), dst = NUT.lladdr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.9b"
	)

	step (2, "We sends an ICMPv6 Echo Request" )
	link1.send(ereq)
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.lladdr , dst = IPv6Address(PRFX2.replace("::","") + "::1") ) / ICMPv6EchoReply() , dst=LBR1.hwaddr	)

	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.9c
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_9c(NUT , TIMER1 , LBR1 , LR1, PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 )

	ereq1 = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.9c-LBR1"
	)
	ereq2 = pack (
		IPv6 (src = LR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.9c-LR1"
	)

	step (2, "We sends an ICMPv6 Echo Request" )
	link1.send(ereq1)
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.addr , dst = LBR1.addr ) / ICMPv6EchoReply() , dst=LBR1.hwaddr )

	set_verdict ("pass")

	step (3, "We sends an ICMPv6 Echo Request" )
	link1.send(ereq2)
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.addr , dst = LR1.addr ) /	ICMPv6EchoReply() , dst=LBR1.hwaddr )

	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.9d
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_9d(NUT , TIMER1 , LBR1, PRFX1 , link1, action):
#FIXME: allow any pl
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 )

	step (2, "NUT action")
	action ("Order the NUT to Send an ICMPv6 Echo Request to %s" % IPV6_ALL_NODES )
	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.addr , dst = IPV6_ALL_NODES ) / ICMPv6EchoRequest() , dst=LBR1.hwaddr )

	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.9e
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_9e(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):
#FIXME: allow any pl
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	addr = IPv6Address( str(PRFX1).replace("::","") + ":FDFF:FFFF:FFFF:FF2A" )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 )

	step (2, "NUT action")
	action ("Order the NUT to Send an ICMPv6 Echo Request to %s" % addr)

	t.start(TIMER1)
	link1.receive (	IPv6( src=NUT.addr , dst = addr) / ICMPv6EchoRequest() , dst=LBR1.hwaddr )

	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.10a
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_10a(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 )

	step (2, "NUT action")
	action ("Order the NUT to sleep for 60 seconds")

	step (3, "Wait 60 seconds")
	t.start (60)
	t.timeout()

	step (4, "We expect the NUT to send an ICMPv6 Neighboor solicitation")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.addr ) /
		ICMPv6NSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			SixARO()
			] )
		)
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.10b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_10b(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1( LBR1 , NUT , PRFX1 , TIMER1, link1 , routerlt=60 , prefixlt=120 )

	step (2, "NUT action")
	action ("Order the NUT to sleep for 60 seconds")

	step (3, "Wait 60 seconds")
	t.start (60)
	t.timeout()

	step (4, "We expect the NUT to send an ICMPv6 Router solicitation")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.addr) /
		ICMPv6RSol( )
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.10c
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_10c(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1b" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1 , routerlt=120 , prefixlt=60 )

	step (2, "NUT action")
	action ("Order the NUT to sleep for 60 seconds")

	step (3, "Wait 60 seconds")
	t.start (60)
	t.timeout()

	step (4, "We expect the NUT to send an ICMPv6 Router solicitation")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.addr) /
		ICMPv6RSol( )
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.10d
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_10d(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	step (1, "We perform the common test setup 1.1" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1 , routerlt=120 , prefixlt=120 , contextlt=60 )

	step (2, "NUT action")
	action ("Order the NUT to sleep for 60 seconds")

	step (3, "Wait 60 seconds")
	t.start (60)
	t.timeout()

	step (4, "We expect the NUT to send an ICMPv6 Router solicitation")
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.addr) /
		ICMPv6RSol( )
	)
	set_verdict ("pass")

################################################################################
# Test_LPND_1.1.11a
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_11a(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq1 = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.11a-1"
	)

	ereq2 = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.11a-b"
	)

	step (1, "We perform the common test setup 1.1b" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1)

	step (2, "We sends an ICMPv6 Echo Request. We expect a compressed ICMPv6 Echo Reply" )
	link1.send(ereq1)
	t.start(TIMER1)

	link1.receive (
		SixLowpanIPHC() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)
	set_verdict("pass")

	step (3, "We wait for 60 seconds..." )
	t.start (60)
	t.timeout()
	step (4, "We sends an ICMPv6 Echo Request. We expect a uncompressed ICMPv6 Echo Reply" )
	link1.send(ereq2)
	t.start(TIMER1)
	link1.receive (
		SixLowpanIPv6() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)
	set_verdict("pass")

################################################################################
# Test_LPND_1.1.11b
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_11b(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.11b"
	)

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 ) ,
			SixCO( c=1 , clen=64, pf=PRFX1 , lt=120)
			]
		)
	)


	step (1, "We perform the common test setup 1.1b" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1)

	step (2, "We sends an ICMPv6 Echo Request. We expect a compressed ICMPv6 Echo Reply" )
	link1.send(ereq)
	t.start(TIMER1)
	link1.receive (
		SixLowpanIPHC() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)
	set_verdict("pass")

	step (3, "LBR1 sends a router advertisement")
	link1.send(ra)

	step (4, "We wait for 60 seconds..." )
	t.start (60)
	t.timeout()

	step (5, "We sends an ICMPv6 Echo Request. We expect a compressed ICMPv6 Echo Reply" )
	link1.send(ereq)
	t.start(TIMER1)
	link1.receive (
		SixLowpanIPHC() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)
	set_verdict("pass")

################################################################################
# Test_LPND_1.1.11c
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_11c(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.11c"
	)

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 ) ,
			SixCO( c=1 , clen=64, pf=PRFX1 , lt=0)
			]
		)
	)

	step (1, "We perform the common test setup 1.1b" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1)

	step (2, "We sends an ICMPv6 Echo Request. We expect a compressed ICMPv6 Echo Reply" )
	link1.send(ereq)
	t.start(TIMER1)
	link1.receive (
		SixLowpanIPHC() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)

	step (3, "LBR1 sends a router advertisement")
	link1.send(ra)

	step (4, "We sends an ICMPv6 Echo Request. We expect an uncompressed ICMPv6 Echo Reply" )
	link1.send(ereq)
	t.start(TIMER1)
	link1.receive (
		SixLowpanIPv6() /
		IPv6( src=NUT.addr ) /
		ICMPv6ERep()
	)
	set_verdict("pass")

################################################################################
# Test_LPND_1.1.11d
################################################################################

@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_11d(NUT , TIMER1 , LBR1 , LBR2 , PRFX1 , link1, action):
	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ereq1 = pack (
		IPv6 (src = LBR1.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.11d"
	)
	ereq2 = pack (
		IPv6 (src = LBR2.addr, dst = NUT.addr ),
		ICMPv6EchoRequest(),
		"6lowpan-nd-1.1.11d"
	)

	ra = pack (
		IPv6( src = LBR2.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR2.hwaddr ) ,
			SixPI( pf=PRFX1 ) ,
			SixCO( c=2 , clen=64, pf=PRFX1 , lt=60)
			]
		)
	)

	step (1, "We perform the common test setup 1.1b" )
	CommonTestSetup_1_1b( LBR1 , NUT , PRFX1 , TIMER1, link1)

	step (2, "LBR1 sends a router advertisement")
	link1.send(ra)

	step (3, "We sends an two ICMPv6 Echo Request. We expect two compressed ICMPv6 Echo Reply" )
	link1.send(ereq1)
	link1.send(ereq2)
	t.start(TIMER1)
	receive_unordered ( link1 ,
		(
			SixLowpanIPHC( dci=1 , dac = 1 ) /
			IPv6( src=NUT.addr , dst=LBR1.addr ) /
			ICMPv6ERep()
		),
		(
			SixLowpanIPHC( dci=2  , dac = 1 ) / #FIXME: verify
			IPv6( src=NUT.addr , dst=LBR2.addr ) /
			ICMPv6ERep()
		)
	)
	set_verdict("pass")

################################################################################
# Test_LPND_1.1.12a
################################################################################
@testcase(SixLoWPAN_ND)
def Test_LPND_1_1_12a(NUT , TIMER1 , LBR1 , PRFX1 , link1, action):

	t = Timer()

	unexpected_packets.activate( link1 )
	timeout.activate( t )

	ra = pack (
		IPv6( src = LBR1.lladdr , dst = NUT.lladdr ) ,
		ICMPv6RAdv( opt = [
			ICMPv6SLL( hw=LBR1.hwaddr ) ,
			SixPI( pf=PRFX1 ) ,
			]
		)
	)

	step (1, "We wait for a Router Solicitation to Send a Router Advertisement" )
	t.start(TIMER1)
	link1.receive (
		IPv6( src = NUT.lladdr , dst = IPV6_ALL_ROUTERS ) /
		ICMPv6RSol( )
	)

	link1.send(ra)
	t.start(TIMER1)

	step (1, "We except a Neighboor solicitation with well formated Address registration option" )
	link1.receive (
		IPv6( src = NUT.addr , dst = LBR1.lladdr ) /
		ICMPv6NSol( opt = superop ( [
			ICMPv6SLL( hw=NUT.hwaddr ) ,
			SixARO(
				type = 31, #TODO: TB1, will be allocated ,
				#length = 2 , not present in the sixARO description
				st = 0 ,
				rsv = 0 ,
				eui = NUT.eui64
				)
			] )
		)
	)
	set_verdict("pass")



################################################################################
# Main()
################################################################################

if __name__ == "__main__":
	with LoggerGroup ([ConsoleLogger(), HTMLLogger()]) as l:
		Logger.set_default (l)
#		ts = TestSession ([ init, Test_LPND_1_1_8c ])
#		ts.set_config(SixLoWPAN_ND, SixLoWPAN_ND("host"))
#		ts.run()
		run_all_testcases({SixLoWPAN_ND: SixLoWPAN_ND("host")})



from ..common import *


class TD_6LOWPAN_HC_12 (SixlowpanTestCase):
    """
---
TD_6LOWPAN_HC_12:
    cfg: Node-Node
    not: Null
    obj: Check that EUTs correctly handle compressed 6LoWPAN packets (16-bit link-local, hop limit=1 and payload=0)
    pre:
        - Header compression is enabled on both EUT1 and EUT2
        - EUT1 and EUT2 are configured to use EUI-64 addresses
        - EUT1 and EUT2 are configured with a default hop limit of 1
    ref: RFC 6282 section 3; RFC 4944 section 4
    seq:
        -   s:
            - EUT1 initiates an echo request to EUT2's link-local address
            - ICMP payload = 0 bytes, total IPv6 size 35 bytes
            - Hop Limit is 1, no traffic class or flow label is being used
        -   c: EUT1 sends a compressed 6LoWPAN packet containing the Echo
                Request message to EUT2
        -   c: "Dispatch value in 6LowPAN packet is 0011TFxHL"
        -   f: "Hop Limit is 64 and source address begin by fe80::"
        -   f: In IP_HC, TF is 11 and the ecn, dscp and flow label fields are
                compressed away
        -   f: In IP_HC, HLIM (HL) is 10 and the hop limit field is compressed
                away
        -   f: In IP_HC, SAC=0, SAM=01; DAC=0; DAM=01
        -   v: EUT2 receives the Echo Request message from EUT1
        -   c: EUT2 sends a compressed 6LoWPAN packet containing the Echo Reply
                message to EUT1
        -   c: "Dispatch value in 6LowPAN packet is 0011TFxHL"
        -   f: "Hop Limit is 64 and source address begin by fe80::"
        -   f: In IP_HC, TF is 11 and the ecn, dscp and flow label fields are
                compressed away
        -   f: In IP_HC, HLIM (HL) is 10 and the hop limit field is compressed
                away
        -   f: In IP_HC, SAC=0, SAM=01; DAC=0; DAM=01
        -   v: EUT1 receives the Echo Reply message from EUT2
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. warning::
            For the moment, we didn't manage to generate packets with the
            wanted size so we just don't take them into account
        """
        return [
            SixLowpanIPHC(
                pl=All(
                    # Length(IPv6, 35),
                    IPv6(
                        tc=0x00,
                        #fl=0x00099cba,
                        hl=1,
                        pl=ICMPv6EchoRequest(
                            # pl=Length(bytes, 0)
                        )
                    )
                )
            )
        ]

    @classmethod
    @typecheck
    def get_nodes_identification_templates(cls) -> list_of(Node):
        """
        Get the nodes of this test case. This has to be be implemented into
        each test cases class.

        :return: The nodes of this TC
        :rtype: [Node]

        """
        return [
            Node('EUT1', ICMPv6EchoRequest()),
            Node('EUT2', ICMPv6EchoReply())
        ]

    def run(self):
        # NOTE: Should we check the IP adresses to check that it is really the
        #       EUT1 and EUT2?

        # TS 1
        self.match('EUT1', SixLowpanIPHC(pl=IPv6(pl=ICMPv6EchoRequest())))

        # TS 2
        self.match('EUT1', SixLowpanIPHC(dp=0b011))

        # TS 3
        self.match('EUT1', SixLowpanIPHC(
            tf=0b01,
            iecn=0b00,
        ))
        self.match('EUT1', SixLowpanIPHC(pl=IPv6(HopLimit=1)))

        # TS 4
        self.match('EUT1', SixLowpanIPHC(hl=0b01, ihl=Omit()))

        # TS 5
        self.match('EUT1', SixLowpanIPHC(
            sac=False,
            sam=0b01,
            dac=False,
            dam=0b01
        ))

        # TS 6
        # NOTE: Only one sniff file so we can't check that the EUT2 didn't
        #       receive the echo request message

        self.next()

        # TS 7
        self.match('EUT2', SixLowpanIPHC(pl=IPv6(pl=ICMPv6EchoReply())))

        # TS 8
        self.match('EUT2', SixLowpanIPHC(dp=0b011))

        # TS 9
        self.match('EUT2', SixLowpanIPHC(
            tf=0b01,
            iecn=0b00,
        ))

        # TS 10
        self.match('EUT2', SixLowpanIPHC(hl=0b01, ihl=Omit()))

        # TS 11
        self.match('EUT2', SixLowpanIPHC(
            sac=False,
            sam=0b01,
            dac=False,
            dam=0b01
        ))

        # TS 12
        # NOTE: Only one sniff file so we can't check that the EUT2 didn't
        #       receive the echo request message



from ..common import *


class TD_LWM2M_1_INT_220(CoAPTestCase):
    """
---
testcase_id: TD_LWM2M_1.0_INT_220
uri: http://openmobilealliance.org/iot/lightweight-m2m-lwm2m
configuration: LWM2M_CFG_01
objective: 
  - Setting the writable resources of object 1 (Server Object_ID 1) Instance 0 in using JSON data format (11543) and restoring these resources to their initial values
  - This test has to be run for the following resources
  - - Default Minimum Period
  - - Default maximum Period
  - - Disable Timeout
  - - Notification Storing When Disabled or Offline
  - - Binding
pre_conditions: 
  - Device is registred with the Server
  - The current values of the Server Object (ID:1) Instance 0, are saved on the Server
  - The Client supports the Configuration C.3 (A superset of C.1 where server objects implements {Default Minimum PEriod, Default Maximum Period, Disable Timeout} additional resources)  
references: 'OMA-ETS-LightweightM2M-V1_0_1-20170926-A'
sequence:
  - step_id: 'TD_LWM2M_1.0_INT_220_step_01'
    type: stimuli
    node: lwm2m_server
    description:
      - 'LwM2M server sends a WRITE request (CoAP PUT) on server object instance'
      - - Type = 0 (CON)
        - Code = 3 (PUT)

  - step_id: 'TD_LWM2M_1.0_INT_220_step_02'
    type: check
    description:
      - 'Sent POST request contains'
      - - Type=0 and Code=3
        - Non-empty payload = array of Default Minimum Period, Default Maximum Period, Disable Timeout, Notification Storing When Disabled or offline, and Binding resource values
        - content-format=application/vnd.oma.lwm2m+json
        - URI-Path option= 1/0

  - step_id: 'TD_LWM2M_1.0_INT_220_step_03'
    type: check
    description:
      - 'LwM2M client sends response containing'
      - - Code = 2.04 (Changed)
        - content-format=application/vnd.oma.lwm2m+json
        - Non-empty payload= A serialized representation of Default Minimum Period, Default Maximum Period, Disable Timeout, Notification Storing When Disabled or offline, and Binding resource values

  - step_id: 'TD_LWM2M_1.0_INT_220_step_04'
    type: verify
    node: lwm2m_server
    description:
      - 'LwM2M server indicates successful operation'

  - step_id: 'TD_LWM2M_1.0_INT_220_step_05'
    type: stimuli
    node: lwm2m_server
    description:
      - 'LwM2M server sends a READ request (CoAP GET) on server object instance'
      - - Type = 0 (CON)
        - Code = 1 (GET)

  - step_id: 'TD_LWM2M_1.0_INT_220_step_06'
    type: check
    description:
      - 'Sent GET request contains'
      - - Type=0 and Code=1
        - Accept option=application/vnd.oma.lwm2m+json
        - URI-Path option= 1/0

  - step_id: 'TD_LWM2M_1.0_INT_220_step_07'
    type: check
    description:
      - 'LwM2M client sends response containing'
      - - Code = 2.05 (Content)
        - content-format=application/vnd.oma.lwm2m+json
        - Non-empty payload
          - A serialized representation of server object instance with Default Minimum Period, Default Maximum Period, Disable Timeout, Notification Storing When Disabled or offline, and Binding resource values
          - The received values are consistent with the 220-SetOfValues sample

  - step_id: 'TD_LWM2M_1.0_INT_220_step_08'
    type: verify
    node: lwm2m_server
    description:
      - 'LwM2M server indicates successful operation'

  - step_id: 'TD_LWM2M_1.0_INT_220_step_09'
    type: stimuli
    node: lwm2m_server
    description:
      - 'LwM2M server sends a WRITE request (CoAP POST) on server object instance'
      - - Type = 0 (CON)
        - Code = 3 (PUT)

  - step_id: 'TD_LWM2M_1.0_INT_220_step_10'
    type: check
    description:
      - 'Sent PUT request contains'
      - - Type=0 and Code=3
        - Non-empty payload = array of initial values (which have been preserved) of Minimum Period, Default Maximum Period, Disable Timeout, Notification Storing When Disabled or offline, and Binding resources
        - content-format=application/vnd.oma.lwm2m+json
        - URI-Path option= 1/0

  - step_id: 'TD_LWM2M_1.0_INT_220_step_11'
    type: check
    description:
      - 'LwM2M client sends response containing'
      - - Code = 2.04 (Changed)
        - content-format=application/vnd.oma.lwm2m+json
        - Non-empty payload

  - step_id: 'TD_LWM2M_1.0_INT_220_step_12'
    type: verify
    node: lwm2m_server
    description:
      - 'LwM2M server indicates successful operation'

  - step_id: 'TD_LWM2M_1.0_INT_220_step_13'
    type: stimuli
    node: lwm2m_server
    description:
      - 'LwM2M server sends a READ request (CoAP GET) on server object instance'
      - - Type = 0 (CON)
        - Code = 1 (GET)

  - step_id: 'TD_LWM2M_1.0_INT_220_step_14'
    type: check
    description:
      - 'Sent GET request contains'
      - - Type=0 and Code=1
        - Accept option=application/vnd.oma.lwm2m+json
        - URI-Path option= 1/0

  - step_id: 'TD_LWM2M_1.0_INT_220_step_15'
    type: check
    description:
      - 'LwM2M client sends response containing'
      - - Code = 2.05 (Content)
        - content-format=application/vnd.oma.lwm2m+json
        - Non-empty payload
          - A serialized representation of server object instance with Default Minimum Period, Default Maximum Period, Disable Timeout, Notification Storing When Disabled or offline, and Binding resource values
          - The received values are consistent with the values present in the initial configuration 3.

  - step_id: 'TD_LWM2M_1.0_INT_220_step_16'
    type: verify
    node: lwm2m_server
    description:
      - 'LwM2M server indicates successful operation'

    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [CoAP(type='con', code='put'), CoAP(type='con', code='get'), CoAP(type='con', code='put')]

    def run(self):
        self.match('server', CoAP(type='con', code='put', opt=self.uri('/1/0')), 'fail')
        self.match('server', CoAP(pl=Not(b'')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionAccept('11543'))), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat('11543'))), 'fail')
        
        self.next()

        self.match('client', CoAP(code=Any(65, 68), pl=(b'')), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat('11543'))), 'fail')

        self.next()

        self.match('server', CoAP(type='con', code='get', opt=self.uri('/1/0')), 'fail')
        self.match('server', CoAP(pl=(b'')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionAccept('11543'))), 'fail')
        
        self.next()

        self.match('client', CoAP(code=2.05, pl=Not(b'')), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat('11543'))), 'fail')

        validation = validate(str(self.coap['pl']),'1')
        self.set_verdict(validation[0], validation[1])

        self.next()

        self.match('server', CoAP(type='con', code='put', opt=self.uri('/1/0')), 'fail')
        self.match('server', CoAP(pl=Not(b'')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionAccept('11543'))), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionContentFormat('11543'))), 'fail')
        
        self.next()

        self.match('client', CoAP(code=Any(65, 68), pl=(b'')), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat('11543'))), 'fail')

        self.next()

        self.match('server', CoAP(type='con', code='get', opt=self.uri('/1/0')), 'fail')
        self.match('server', CoAP(pl=(b'')), 'fail')
        self.match('server', CoAP(opt=Opt(CoAPOptionAccept('11543'))), 'fail')
        
        self.next()

        self.match('client', CoAP(code=2.05, pl=Not(b'')), 'fail')
        self.match('client', CoAP(opt=Opt(CoAPOptionContentFormat('11543'))), 'fail')

        validation = validate(str(self.coap['pl']),'1')
        self.set_verdict(validation[0], validation[1])



#!/usr/bin/env python3

from ..common import *
from ttproto.tat_coap.testcases.td_coap_core_18 import TD_COAP_CORE_18


class TD_COAP_CORE_24 (TD_COAP_CORE_18):
    """
---
TD_COAP_CORE_24:
    cfg: CoAP_CFG_03
    obj: Perform POST transaction with responses containing several
         Location-Path options (Reverse Proxy in CON mode)
    pre:
        - Proxy is configured as a reverse-proxy for the server
        - Proxy’s cache is cleared
        - Server accepts creation of new resource on /create2 and the created
          resource is located at /location1/location2/location3 (resource does
          not exist yet)
    ref: '[1] clause 5.8.1,5.10.8,5.9.1.1, 8.2.2,8.2.1,10.2.2,11.2'
    seq:
        -   s: Client is requested to send a confirmable POST request to proxy
        -   c:
            - The POST sent by the client contains:
            -   - Type = 0 (CON)
                - Code = 2 (POST)
                - An arbitrary payload
                - Content-format option
        -   c:
            - 'The Proxy forwards the POST request to server’s resource and
               that it contains:'
            -   - Type = 0 (CON)
                - Code = 2 (POST)
                - An arbitrary payload
                - Content-format option
        -   c:
            - 'Server sends a response to the proxy containing:'
            -   - Code = 65 (2.01 created)
                - Option type = Location-Path  (one for each segment)
                - 'Option values must contain "location1", "location2" &
                   "location3" without contain any "/"'
        -   c:
            - 'Observe that the Proxy forwards the response (in step 4) to
               client and check that the forwarded response contains:'
            -   - Code = 65 (2.01 created)
                - Option type = Location-Path  (one for each segment)
                - 'Option values must contain "location1", "location2" &
                   "location3" without contain any "/"'
        -   v: Client displays the response
        -   v:
            - Client interface returns the response
            -   - 2.01 created
                - Location: coap://proxy/location1/location2/location3
    """

    request_uri = "/create2"
    reverse_proxy = True



import unittest
import os
import json
import logging

from ttproto.core.dissector import Capture, get_dissectable_protocols
from ttproto.core.packet import PacketValue
from ttproto.core.lib.inet.sixlowpan import SixLowpan
from ttproto.utils.pcap_filter import openwsn_profile_filter

from tests.test_tools.struct_validator import StructureValidator
from tests.test_dissector.test_dissect_pcaps import DissectPcapTestCase


class DissectorTestCase_6lowpan(DissectPcapTestCase):
    """
    Test class for the dissector tool

    python3 -m unittest tests.test_dissector.test_dissector_6lowpan.DissectorTestCase.test_that_it_can_dissect_all_pcaps -vvv
    """

    # #################### Tests parameters #########################

    # File path
    PCAP_FILES_DISSECTION_DIRS = ['tests/test_dumps/6lowpan',
                                  'tests/test_dumps/6lowpan_hc',
                                  'tests/test_dumps/6lowpan_nd']
    PROTO_CLASS_FILTER = SixLowpan
    TMP_DIR = 'tmp/'

    # Create a struct checker object
    struct_validator = StructureValidator()



# # #################### Main run the tests #########################
if __name__ == '__main__':
    unittest.main(verbosity=3)

    # 15.4 complete file
    # self.PCAP_FILE_LINKTYPE_IEEE802_15_4_FCS = self.PCAP_FILES_DISSECTION + os.sep + "wpan_802_15_4_LLT_195_6lowpan.pcap"
    # self.pcap_for_test = []
    # self.pcap_for_test.append(self.PCAP_FILE_LINKTYPE_IEEE802_15_4_FCS)
    # self.filtered_pcap_filename = openwsn_profile_filter(self.PCAP_FILE_LINKTYPE_IEEE802_15_4_FCS)
    # self.dissector = Dissector(self.filtered_pcap_filename)

    # 15.4 ACKS
    # self.dissector = Dissector(self.PCAP_FILE_LINKTYPE_IEEE802_15_4_ACKS)
    # self.filtered_pcap_filename = openwsn_profile_filter(self.PCAP_FILE_LINKTYPE_IEEE802_15_4_ACKS)
    # self.dissector = Dissector(self.filtered_pcap_filename)

    # CoAP
    # PCAP_FILES_DISSECTION = 'tests/test_dumps/dissection'
    # PCAP_FILE = PCAP_FILES_DISSECTION + '/CoAP_plus_random_UDP_messages.pcap'
    # self.dissector = Dissector(PCAP_FILE)

    # 15.4 FCS

    # PCAP_FILE_LINKTYPE_IEEE802_15_4_FCS= 'tests/test_dumps/6lowpan/802_15_4_with_FCS_13_frames.pcap'
    # self.dissector = Dissector(PCAP_FILE_LINKTYPE_IEEE802_15_4_FCS)
    #
    # self.dissector = Dissector(self.filtered_pcap_filename)

    # 802_15_4_simple_FCS_test
    # pcap_name = "802_15_4_simple_FCS_test.pcap"
    # self.pcap_for_test=[]
    # file = os.path.join(self.PCAP_FILES_DISSECTION,"openwsn_captures",pcap_name)
    # self.filtered_pcap_filename = openwsn_profile_filter(file,'filtered_'+ pcap_name)
    # self.pcap_for_test.append(

    # #################### Utilities functions #########################
    #
    # def test_check_summary(self, summary):
    #     self.assertTrue(type(summary), tuple)
    #     self.assertEqual(len(summary), 2)
    #     self.assertTrue(type(summary[0]), int)
    #     self.assertGreater(summary[0], 0)
    #     self.assertTrue(type(summary[1]), str)
    #     self.assertGreater(len(summary[1]), 0)

    # #################### Tests functions #########################

    # ##### get_implemented_protocols



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

"""The core.control module provides the function for handling the testcase
definition and for their execution.
"""


import collections, gc, signal, threading, time, traceback, types, os

from	ttproto.core.typecheck	import *
from	ttproto.core		import clock, data, exceptions, logger, snapshot

__all__ = [
	'all_testcases',
	'Verdict',
	'VerdictValue',
	'VerdictSummary',
	'init_verdict',
	'set_verdict',
	'get_verdict',
	'Config',
	'Testcase',
	'EventTestcaseStarted',
	'EventTestcaseTerminated',
	'EventTestcaseRuntimeError',
	'TestSession',
	'EventTestSessionStarted',
	'EventTestSessionAborted',
	'EventTestSessionError',
	'EventTestSessionTerminated',
	'EventSetVerdict',
	'testcase',
	'run_all_testcases',
	'stop',
	'SUTAction',
	'SUTActionFailed',
	'SUTActionManual',
]


all_testcases = []

__thread_local = threading.local()

class Verdict:
	"""A class handling the verdict for a test case

	Known verdict values are:
	 - "none"	no verdict set yet
	 - "pass"	the NUT fulfilled the test purpose
	 - "inconclusive"	the NUT did not fulfill the test purpose but did not display bad behaviour
	 - "fail"	the NUT did not fulfill the test purpose and displayed a bad behaviour
	 - "aborted"	the test execution was aborted by the user
	 - "error"	a runtime error occured during the test


	At initialisation time, the verdict is set to None. Then it can be updated one or
	multiple times, either explicitely calling set_verdict() or implicitely if an
	unhandled exception is caught by the control module (error verdict) or if the user
	interrupts the test manually (aborted verdict).

	Each value listed above has precedence over the previous ones. This means
	that when a verdict is updated, the resulting verdict is changed only if
	the new verdict is worse than the previous one.
	"""

	__values = ("none", "pass", "inconclusive", "fail", "aborted", "error")

	def __init__ (self, initial_value = None):
		self.__value = 0
		if initial_value is not None:
			self.update (initial_value)

	@typecheck
	def update (self, new_verdict:str , message: str = ""):
		"""Update the verdict"""

		assert new_verdict in self.__values

		new_value = self.__values.index(new_verdict)
		if new_value > self.__value:
			self.__value = new_value

		logger.log_event (EventSetVerdict (new_verdict, message))

	@classmethod
	def values (cls):
		"""List the known verdict values"""
		return cls.__values

	def get_value (self):
		"""Get the value of the verdict"""
		return self.__values [self.__value]

	def __str__ (self):
		return self.__values [self.__value]

VerdictValue = str

class VerdictSummary:
	"""A class holding the list of the testcases that are run, with their verdict"""
	def __init__ (self):
		self.__summary = collections.Counter()
		for k in Verdict.values():
			self.__summary[k] = 0

		self.__testcase_results = []

	@typecheck
	def update (self, testcase, verdict: VerdictValue):
		"""Update the summary with a new test result"""
		assert isinstance (testcase, Testcase)

		self.__summary.update ((verdict,))
		self.__testcase_results.append ((testcase, verdict))

	def count (self):
		"""return the number of testcase that were run"""
		nb = 0
		for v in self.__summary.values():
			nb += v
		return nb

	def __iter__ (self):
		"""Return an iterator of an histogram of the verdict values

		The iterator yields a set of tuples “(verdict, count)” indicating
		the occurence of each verdict.
		"""
		for v in Verdict.values():
			yield v, self.__summary[v]

	def __str__ (self):
		return ", ".join ("%s: %d" % (k, self.__summary[k]) for k in Verdict.values())

	def testcases (self):
		"""Return an iterator on the run testcases with their verdict

		The iterator yields a set of tuples “(testcase, verdict)”
		"""
		return iter (self.__testcase_results)

def init_verdict():
	"""Initialise the verdict of the current thread"""
	threading.current_thread().verdict = Verdict()

# TODO: log the message
def set_verdict(new_verdict, message = ""):
	"""Update the verdict of the current thread"""
	threading.current_thread().verdict.update (new_verdict)

def get_verdict():
	"""Get the verdict of the current thread"""
	return threading.current_thread().verdict.get_value()


class Config:
	"""A class for parametering the test cases and for setting up the test configuration.

	A configuration can contain one on more named parameters that are
	stored in a dict and used as actual parameters when running the the
	testcases.

	The Config class can also be used to handle test setup and cleanup
	(through the __enter__ and __exit__ functions).

	Example:
		class TestConfig (Config):
			pass

		@testcase (TestConfig)
		def TC_config_test (A, B, C):
			print (A, B, C)
			set_verdict ("pass")



		my_cfg = TestConfig ({"A": 12, "B": 48})

		my_cfg["B"] = 27
		my_cfg["C"] = 137

		run_all_testcases ({TestConfig: my_cfg})

	Output:
		[795.5007] Testcase Started:    TC_config_test
		12 27 137
		[795.5008] Testcase Terminated: TC_config_test --> pass
	"""
	#TODO: use a metaclass to do some typechecking on the parameters and on the testcase parameters (like in LogEvents)

	@typecheck
	def __init__ (self, dict_init_value: iterable = ()):
		"""Initialise the configuration

		The dict_init_value, if provided contains the intialisation
		value for the dictionnary of parameters.
		"""
		self.__params = dict (dict_init_value)

	def run (self, testcase, func, **kw):
		"""Run a testcase main function using this configuration

		This function enter the self context (“with self:”) and calls
		the function 'func' using the parameters provided in **kw.
		Before calling the function, it does some introspection in
		'func' to determine if this function requires other parameters
		(not provided by **kw), to fill them with the parameters set in
		the internal dict of this Config object (or raise an error if
		the parameter is unknown).

		In case of runtime error (unhandled exception) the function
		catches the exception and generates an
		EventTestcaseRuntimeError log event describing the error.
		"""
		try:
			stage = 0
			with self:
				stage = 1
				for i in range (0, func.__code__.co_argcount):
					name = func.__code__.co_varnames[i]

					if name not in kw:
						if name in self.__params:
							kw[name] = self.__params[name]
						else:
							raise exceptions.Error ("Missing configuration parameter '%s'" % name)
				stage = 2
				func (**kw)
				stage = 3

		except exceptions.UserInterrupt:
			raise

		except exceptions.TerminateTestcase:
			pass

		except Exception as e:
			if stage < 2:
				msg = "configuration setup failed (%s)"
			elif stage > 2:
				msg = "configuration cleanup failed (%s)"
			else:
				msg = "%s"

			logger.log_event (EventTestcaseRuntimeError (testcase, msg % str(e), traceback.format_exc()))
			set_verdict("error")

	def __setitem__ (self, key, value):
		"""Set one config parameter"""
		self.__params[key] = value

	def __getitem__ (self, key):
		"""Get one config parameter"""
		return self.__params[key]

	def __enter__ (self):
		"""Enter this config context"""
		pass

	def __exit__ (self, a, b, c):
		"""Leav this config context"""
		pass


class Testcase (snapshot.EventSource):
	"""A class representing a testcase

	A testcase is associated to a function (the entry point of the
	testcase) and a type of configuration (subclass of Config) that must be
	provided at runtime so that the test can run properly.

	Testcase objects are usually not instantiated directly, but with the
	help @testcase decorator function.
	"""

	def __init__ (self, func, config_type: type = Config):
		"""Initialise the testcase

		- 'func' is the main function of the testcase
		- 'config_type' is the type of config that must be provided at
		  runtime so that the testcase can run properly (by default
		  Config is used)

		The name of the testcase (__name__ attribute is initialised
		from the name of the function 'func').
		"""
		assert issubclass (config_type, Config)

		super().__init__()

		self.__func = func
		self.__name__ = func.__name__
		self.__config_type = config_type
		self.__terminated = False
		self.__snapshot_terminated = False

		self.terminated = snapshot.Event (self, "terminated")

	def evaluate_snapshot (self):
		self.__snapshot_terminated = self.__terminated

	def match_terminated (self):
		return snapshot.EventMatch (self.terminated) if self.__snapshot_terminated else None

	def get_config_type (self):
		"""Return the type of config required by this testcase"""
		return self.__config_type

	def run (self, config):
		"""Runs the testcase.

		This function starts a new thread to run the testcase function.
		Once the testcase is terminated, it returns the final verdict.

		It also generate log events EventTestcaseStarted and
		EventTestcaseTerminated.
		"""
		assert isinstance (config, self.__config_type)

		th = snapshot.SnapshotManager.Thread (target = self.__thread_func, args = (config,))

		th.start()

		#FIXME: what happens if we are not the main thread (we'll not catch the SIGINT signals)
		self.terminated()

		return th.verdict.get_value()

	def __thread_func (self, config):
		init_verdict()

		logger.log_event (EventTestcaseStarted (self))
		try:
			with data.Data.disable_name_resolution(): # because it is very time consuming
				config.run (self, self.__func)
		except exceptions.UserInterrupt as e:
			set_verdict("aborted")
		except Exception as e:
			traceback.print_exc()
			set_verdict("error")

		logger.log_event (EventTestcaseTerminated (self, get_verdict()))

		with self.lock:
			self.__terminated = True

		gc.collect()

class EventTestcaseStarted (metaclass = logger.LogEventClass):
	fields = (("testcase", Testcase),)
	def summary (self):
		return "Testcase Started:    %s" % (self[0].__name__)

class EventTestcaseTerminated (metaclass = logger.LogEventClass):
	fields = (("testcase", Testcase),
		  ("verdict", VerdictValue),
		 )
	def summary (self):
		return "Testcase Terminated: %s --> %s" % (self[0].__name__, self[1])

class EventTestcaseRuntimeError (metaclass = logger.LogEventClass):
	fields = (("testcase", Testcase),
		  ("message", str),
		  ("traceback", str),
		 )
	def summary (self):
		return "Testcase Runtime Error: %s: %s" % (self[0].__name__, self[1])

class EventSetVerdict (metaclass = logger.LogEventClass):
	fields = (("verdict", VerdictValue),
		  ("message", str),
		 )
	def summary (self):
		return "Verdict: %s%s" % (
				self[0],
				(" - %s" % self[1]) if self[1] else "")

class TestSession (logger.LoggedObject):
	"""A class for handling test sessions

	a test session contains a list of testcases to be run and a set of
	configurations for these testcases.
	"""
	@typecheck
	def __init__ (self, testcase_list: optional (list_of (Testcase)) = None, description: str = ""):
		"""Initialise the test session

		Parameters:
		- testcase_list (opt)	list of testcases to be run (by default, all testcases are run)
		- description (opt)	description of the test session


		"""
		logger.LoggedObject.__init__(self)
		if testcase_list is None:
			self.__tc_list = list (all_testcases)
			self.__description = "all testcases"

		else:
			self.__tc_list = list (testcase_list)
			self.__description = description

		self.__configs = {Config: Config()}

	def get_description (self):
		"""Return the textual description of the test session"""
		return self.__description

	@typecheck
	def set_config (self, cfg_type: type, config: Config):
		"""Set a configuration

		This function stores which config object 'config' shall be used
		for testcases requiring the type of configuration 'cfg_type'.
		"""
		assert issubclass (cfg_type, Config)
		assert isinstance (config, cfg_type)
		self.__configs[cfg_type] = config

	def run (self):
		"""Run the test session"""

		#FIXME: not really necessary to have a separate thread
		th = snapshot.SnapshotManager.Thread (target = self.__thread_func)

		th.start()

		th.interrupt = types.MethodType (self.interrupt, th)
		th.__last_interruption = 0
		th.__interrupted = False

		#FIXME: join() w/o any timeout will not catch ^C interrupts (maybe a bug in python)
		#FIXME: what happens if we are not the main thread (we'll not catch the SIGINT signals)
		th.join(1231322222246514)

	@staticmethod
	def interrupt (thread):
		"""Function for notifying an interruption by the user.

		Once this function is called, the session thread is notified
		that it should abort the execution"""
		t = time.time()
		if (t - thread.__last_interruption) < 1:	# TODO: handle this in the user interactions interface
			thread.__interrupted = True
		thread.__last_interruption = t

	def execute (self, testcase, summary):
		"""Execute one testcase in this session

		This function selects an adequate configuration for the given
		testcase, and runs it (or report an error if the the requested
		configuration type is missing).

		The summary is updated with the outcome of the testcase.

		The function returns the final verdict of the testcase.
		"""

		cfg_type = testcase.get_config_type()

		if cfg_type in self.__configs:
			verdict = testcase.run (self.__configs[cfg_type])
		else:
			logger.log_event (EventTestSessionError (self, "missing configuration: %s" % cfg_type.__name__))
			verdict = "error"

		summary.update (testcase, verdict)

		if threading.current_thread().__interrupted:
			raise exceptions.UserInterrupt()

		return verdict

	def __thread_func (self):

		summary = VerdictSummary()

		self.log (EventTestSessionStarted)
		try:
			for tc in self.__tc_list:
				self.execute (tc, summary)

		except exceptions.UserInterrupt:
			self.log (EventTestSessionAborted)

		self.log (EventTestSessionTerminated, summary)

		gc.collect()

class EventTestSessionStarted (metaclass = logger.LogEventClass):
	fields = (("session", TestSession),)
	def summary (self):
		return "Test session started: %s" % self[0].get_description()

class EventTestSessionAborted (metaclass = logger.LogEventClass):
	fields = (("session", TestSession),)
	def summary (self):
		return "Test session aborted: %s" % self[0].get_description()

class EventTestSessionError (metaclass = logger.LogEventClass):
	fields = (("session", TestSession),
		  ("message", str))
	def summary (self):
		return "Test session error: %s" % self[1]

class EventTestSessionTerminated (metaclass = logger.LogEventClass):
	fields = (("session", TestSession),
		  ("verdict_summary", VerdictSummary))
	def summary (self):
		result = ["Test session terminated: %s\n\nTestcase results:" % self[0].get_description()]
		summary = self[1]

		for tc, v in summary.testcases():
			result.append ("\t[%s]  %s" % (str(v).center(8), tc.__name__))

		result.append("\nSummary:")
		total = summary.count()
		for v, nb in summary:
			result.append ("\t%-10s %3d  (%3d%%)" % (v, nb, nb*100//total))

		result.append ("\n\t%-10s %3d  (%3d%%)" % ("Total", total, 100))

		return "\n".join(result)

@typecheck
def testcase (arg: either (types.FunctionType, type)):
	"""Decorator function for defining testcases

	this function can be used in two ways:

	1. testcases that do not require any specific configuration. the
	testcase configuration will by default use the type config.

		@testcase
		def my_testcase():
			...

	2. testcases that requires a specific configuration. this configuration
	shall be given as parameter.

		@testcase (MyConfig)
		def my_testcase():
			...
	"""
	if isinstance (arg, type):
		assert issubclass (arg, Config)

		def tc_def (func):
			tc = Testcase (func, arg)
			all_testcases.append (tc)
			return tc
		return tc_def

	else:
		tc = Testcase (arg)
		all_testcases.append (tc)
		return tc
@typecheck
def run_all_testcases (configs: dict_of(type, Config) = {}):
	"""Run all the known testcases

	This function instantiates a TestSession with all the known testcases,
	in the order they were defined and runs it.

	The 'configs' parameter may contain a dict of the configuration objects
	to be used for each type of configuration used in the testcases.
	"""
	ts = TestSession()
	for k,v in configs.items():
		ts.set_config (k, v)
	ts.run()

def _sigint_handler (sig, frame):
	"""A handler for the SIGINT signal

	This function notifies the running threads that the user is
	interrupting the execution, so that they can terminate cleanly.
	"""
	# FIXME: it would be better to only send the interrupt signal to the main thread and let it
	#	propagate the signal to other threads (this would be more deterministic)
	for t in threading.enumerate():
		if hasattr (t, "interrupt"):
			t.interrupt()

def stop():
	raise exceptions.TerminateTestcase()

signal.signal (signal.SIGINT, _sigint_handler)

class SUTActionFailed (Exception):
	pass

# FIXME: possible race condition if one action is started when another is matched
# TODO:  support concurrent actions
class SUTAction (snapshot.EventSource, logger.LoggedObject):

	@classmethod
	def async (cls, action):
		return cls (action, async = True)

	def __new__ (cls, action, async = False):
		if async:
			return super().__new__(cls)

		# sync case
		handle = cls (action, True)

		try:
			with snapshot.alt:
				handle.done()

				@handle.error
				def _():
					set_verdict ("error", "SUT action error")
		finally:
			handle.interrupt()

	def __init__ (self, action, async):
		snapshot.EventSource.__init__ (self)
		logger.LoggedObject.__init__ (self)

		self.done  = snapshot.Event (self, "done")
		self.error = snapshot.Event (self, "error")

		with self.lock:
			self.action = action
			self.__status = None
			self.__thread = threading.Thread (target = self.__run)
			self.__thread.start()

	def __run (self):
		try:
			self.run()
			result = True
		except Exception as e:
			result = e

		with self.lock:
			self.__status = result
			self.__thread = None

	def run (self):
		raise NotImplementedError()

	def interrupt (self):
		raise NotImplementedError()

	def evaluate_snapshot (self):
		self.__snapshot_status = self.__status

	@typecheck
	def match_done (self) -> optional (snapshot.EventMatch):
		if not self.__snapshot_status:
			return None

		with self.lock:
			self.__status = None

		self.log (EventSUTActionDone, str (self.action))

		return snapshot.EventMatch (self.done)

	@typecheck
	def match_error (self) -> optional (snapshot.EventMatch):
		if not isinstance (self.__snapshot_status, Exception):
			return None

		with self.lock:
			status = self.__status
			self.__status = None

		self.log (EventSUTActionError, str (self.action), status)

		return snapshot.EventMatch (self.error)

class SUTActionManual (SUTAction):

	def __init__ (self, *k):
		self.__pid = None
		super().__init__(*k)

	def run (self):
		with self.lock:
			self.__pid = os.fork()

		if not self.__pid:
			os.execl ("/usr/bin/zenity", "zenity", "--info", "--title=TTProto", "--text=SUT informal action requested:\n\n%s" % self.action)
			print('error: execl("/usr/bin/zenity", ...) failed')
			sys.exit (1)

		pid, status = os.waitpid (self.__pid, 0)

		with self.lock:
			self.__pid = None

		if status // 256:
			raise SUTActionFailed()

	def interrupt (self):
		with self.lock:
			if self.__pid:
				os.kill (self.__pid, signal.SIGTERM)

class EventSUTActionDone (metaclass = logger.LogEventClass):
	fields = (("driver", SUTAction),
		  ("action", str),
		 )

	def summary (self):
		return "SUT action done (%s): %s" % (type(self[0]).__name__, self[1])

class EventSUTActionError (metaclass = logger.LogEventClass):
	fields = (("driver", SUTAction),
		  ("action", str),
		  ("exception", str),
		  ("traceback", str),
		 )

	def summary (self):
		return "SUT action error (%s): %s -> %s" % (type(self[0]).__name__, self[1], repr(self[2]))

#@testcase
#def hello_world():
#	"""Sample hello world test case"""
#	print ("Hello World!")
#	set_verdict ("pass")




from ..common import *


class TD_COAP_OBS_10(CoAPTestCase):
    """
---
TD_COAP_OBS_10:
    obj: GET does not cancel resource observation
    cfg: CoAP_CFG_BASIC
    ref: '[OBSERVE] 1.2, 3, 4'
    pre:
        - Client supports Observe option
        - Server supports Observe option
        - "Server offers an observable resource /obs which changes periodically
            (e.g. every 5s) which produces confirmable notifications"

    seq:
        -   c: "Client is requested to send to the server a confirmable
                GET request with observe option for resource /obs"

        -   c:
            - "The request sent by client contains:"
            -   - Type = 0 (CON)
                - Code = 1 (GET)
                - Token value  = a value generated by the client
                - Observe option = empty

        -   c:
        - "Server sends the response containing:"
        -   - Type = 2 (ACK)
            - Code = 2.05 (Content)
            - Content-format of the resource /obs
            - Token value = same as one found in the step 2
            - Observe option with a sequence number

        -   c:
        - "Server sends a notification containing:"
        -   - Type = 0 (CON)
            - Code = 2.05 (Content)
            - Content-format = same as one found in the step 3
            - Token value = same as one found in the step 2
            - Observe option indicating increasing values

        -   c: Client sends an ACK

        -   v: Client displays the received information

        -   s: "Client is requested to send to the server a confirmable
                GET request *without* observe option for resource /obs"

        -   c:
        - "Client sends a request containing :"
        -  - Type = 0 (CON)
            - Code = 1 (GET)
            - Token value ➔ t2, a value generated by the client ≠ t
            - "*No* Observe option"

        -   c:
        -   - Type = 2 (ACK)
            - Code = 2.05 (Content)
            - Content-format of the resource /obs, = f
            - Token value = t2, same as one found in the step 8
            - "*No* Observe option"

        -   c:
        -   - Type = 0 (CON)
            - Code = 2.05 (Content)
            - Content-format = f, same as one found in the step 3
            - Token value = t, same as one found in the step 3
            - Observe option indicating increasing values

        -   c: Client sends an ACK

        -   v: Client displays the received information

    """
    # "OBS 10 : Consolidated test case and added required stimulis"
    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]
        """
        return [
                CoAP(type='con', code='get', opt=Opt(CoAPOptionObserve(0),CoAPOptionUriPath("obs"))),
                CoAP(type='con', code='get', opt=All(Opt(CoAPOptionUriPath("obs")),
                                                    NoOpt(CoAPOptionObserve())))
        ]

    def run(self):
        # Step 2
        self.match("client", CoAP(type="con", code="get",
                    opt=self.uri("/obs", CoAPOptionObserve(0))))

        token = self.coap["tok"]
        uri = self.coap.get_uri()
        path = self.coap["opt"][CoAPOptionUriPath]["val"]

        self.next()

        # Step 3
        if not self.match("server", CoAP(type="ack", code=2.05, opt=Opt(
            CoAPOptionObserve(),
            CoAPOptionContentFormat(),
        ))):
            raise self.Stop()
        self.match("server", CoAP(tok = token), "fail")

        content_format = self.coap["opt"][CoAPOptionContentFormat]["val"]
        index          = self.coap["opt"][CoAPOptionObserve]["val"]

        self.next()
        # we need to observe at least one notification
        # (to ensure that the server is observing the resource for the client)

        lastFrame = None

        verdict_if_none = "inconclusive"
        while self.match("server", CoAP(code=2.05, opt=Opt(CoAPOptionObserve())),
                        verdict_if_none):
            # Step 4
            if not self.match("server", CoAP(type="con", code=2.05, opt=Opt(
                CoAPOptionObserve(),
                CoAPOptionContentFormat(),
            ))):
                raise self.Stop()

            self.match("server", CoAP(tok = token), "fail")
            self.match("server", CoAP(opt = Opt(CoAPOptionContentFormat(content_format))), "fail")

            new_index = self.coap["opt"][CoAPOptionObserve]["val"]

            self.set_verdict(("pass" if new_index > index else "fail"),
                "value of observe option must be increasing")

            self.next(True)

            # Step 5
            self.match("client", CoAP(type="ack", code=0))

            self.next(True)
            # now we have successfully observed a observe response
            verdict_if_none = None
            if lastFrame == self._frame:
                self.set_verdict("inconclusive",
                    "The stimulus 'Perform an unrelated GET' did not happened\
 or is not visible in any frame this test case operate on")
                raise self.Stop()
            lastFrame = self._frame

        # Step 8
        if self.match("client", CoAP(type="con", code="get"
                                , opt=All(Opt(CoAPOptionUriPath(path)),
                                NoOpt(CoAPOptionObserve()))
                                ) ,None):
            self.set_verdict("pass", "Client performed an unrelated GET.")
            self.match("client", CoAP (tok = Not(token)),"fail")
            token2 = self.coap["tok"]

            self.next()
            # Step 9
            self.match("server", CoAP(type="ack", code=2.05, opt=Opt(
                    CoAPOptionContentFormat(content_format),
                )),"fail")
            self.match("server", CoAP(tok = token2, opt=NoOpt(CoAPOptionObserve())), "fail")

        # Step 10

        while self.match("server", CoAP (type="con",code=2.05, opt=Opt(CoAPOptionObserve(),CoAPOptionContentFormat())), None):
                self.next()
                self.match("server", CoAP(tok = token), "fail")
                self.match("server", CoAP(opt = Opt (CoAPOptionContentFormat(content_format))), "fail")

                new_index = self.coap["opt"][CoAPOptionObserve]["val"]

                self.set_verdict(("pass" if new_index > index else "fail"),
                    "value of observe option must be increasing")

                self.next()

                # Step 11
                if self.match("client", CoAP(type="rst", code=0), None):
                    rst_sent_by_client = True
                    self.set_verdict("pass", "End of communication from the client.")
                    raise self.Stop()
                elif self.match("client", CoAP(type="ack", code=0, tok = token), None):
                    self.set_verdict("pass", "ACK from client.")
                    self.next()

        raise self.Stop()



#!/usr/bin/env python3
#
#  (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re

from ttproto.core.typecheck import *
from ttproto.core.data import *
from ttproto.core.list import ListValue

__all__ = [
    'Range',
    'ValueList',
    'AnyValue',
    'Length',
    'Superset',
    'Regex',
    'All',
    'Any',
    'Not'
]


class Range(Template):
    @typecheck
    def __init__(
        self,
        parent: either(is_data, is_type),
        lower_bound: is_flat_value,
        upper_bound: is_flat_value
    ):

        Template.__init__(self, parent)
        self.__lower = self.store_data(lower_bound)
        self.__upper = self.store_data(upper_bound)

    def _template_match(self, value):
        return self.__lower <= value <= self.__upper

    def _repr(self):
        return "%s(%s, %s, %s)" % (
            type(self).__name__,
            self.get_type().__name__,
            repr(self.__lower),
            repr(self.__upper),
        )


# TODO: rename it as Or
class ValueList(Template):

    @typecheck
    def __init__(self, parent: either(is_type, is_data), datas: iterable):

        Template.__init__(self, parent)

        self.__datas = [self.store_data(d) for d in datas]

    def _template_match(self, value):

        for d in self.__datas:
            if d.match(value):
                return True

        return False

    def _repr(self):
        return "%s(%s, [%s])" % (
            type(self).__name__,
            self.get_type().__name__,
            ", ".join(repr(v) for v in self.__datas),
        )


class AnyValue(Template):
    @typecheck
    def __init__(self, type_: is_type):
        Template.__init__(self, type_)

    def _template_match(self, value):
        return True

    def _repr(self):
        return "%s(%s)" % (
            type(self).__name__,
            self.get_type().__name__,
        )


class Length(Template):
    @typecheck
    def __init__(
        self,
        parent: either(is_type, is_data),
        length: (either(
            int,
            tuple_of(int),
            list_of(either(
                int,
                tuple_of(int)))
            ))
    ):
        Template.__init__(self, parent)

        if not isinstance(length, list):
            length = [length]

        self.__length = [
            (v if isinstance(v, tuple) else(v, v))
            for v in length
        ]

        # check that 'length' contains valid values
        for lower, upper in self.__length:
            assert lower >= 0
            assert upper >= lower

    def _template_match(self, data):
        l = len(data)
        for lower, upper in self.__length:
            if lower <= l <= upper:
                return True
        return False

    def __initlist(self):
        for l, u in self.__length:
            if l == u:
                yield str(l)
            else:
                yield "(%s, %s)" % (l, u)

    def _repr(self):
        return "%s(%s, %s)" % (
            type(self).__name__,
            self.get_type().__name__,
            (
                "[%s]"
                %
                (", ".join(self.__initlist()))
                if len(self.__length) > 1
                else next(self.__initlist())
            )
        )


class Superset(Template):
    @typecheck
    def __init__(self, parent, datas: iterable):

        Template.__init__(self, parent)

        assert issubclass(self.get_type(), ListValue)
        assert not self.get_type().is_ordered()
        assert all(is_data(d) for d in datas)

        ct = self.get_type().get_content_type()
        self.__datas = tuple(store_data(d, ct) for d in datas)

    def _repr(self):
        return "%s(%s, [%s])" % (
            type(self).__name__,
            self.get_type().__name__,
            ", ".join(repr(v) for v in self.__datas),
        )

    def _template_match(self, value):
        missing = len(value) - len(self.__datas)
        if missing < 0:
            return False
        else:
            # TODO: do it in a more efficient way
            return self.get_type()(
                self.__datas + missing * (
                    AnyValue(self.get_type().get_content_type()),)
                ).match(value)

        return False


class Regex(Template):
    @typecheck
    def __init__(self, pattern: str, flags: optional(int) = 0):

        Template.__init__(self, str)

        self.__re = re.compile(pattern, flags)

    def _repr(self):
        flags = self.__re.flags
        flist = []
        for flagname in (
            'ASCII', 'DEBUG', 'IGNORECASE', 'LOCALE',
            'MULTILINE', 'DOTALL', 'VERBOSE', 'UNICODE'
        ):
            v = getattr(re, flagname)
            if v & flags:
                flags &= ~v
                flist.append("re." + flagname)
        if flags:
            flist.append(str(flags))

        return "%s(%s, %s)" % (
            type(self).__name__,
            repr(self.__re.pattern),
            "|".join(flist),
        )

    def _template_match(self, value):
        return bool(re.search(self.__re, value))


class All(Template):
    @typecheck
    def __init__(self, *templates):
        assert templates

        Template.__init__(self, get_type(templates[0]))  # FIXME: bof

        self.__templates = [self.store_data(t) for t in templates]

    def _repr(self):
        return "%s(%s)" % (
            type(self).__name__,
            ",".join(repr(t) for t in self.__templates),
        )

    # here we override _match since we need to match all templates
    #
    # FIXME: may have side effects in the HTML log module(we may have
    # multiple mismatch for the same value)
    def _match(self, value, mismatch_list):
        result = True
        for t in self.__templates:
            if not t.match(value, mismatch_list):
                result = False
        return result


class Any(Template):
    @typecheck
    def __init__(self, *templates):
        assert templates

        Template.__init__(self, get_type(templates[0]))  # FIXME: bof

        self.__templates = [self.store_data(t) for t in templates]

    def _repr(self):
        return "%s(%s)" % (
            type(self).__name__,
            ",".join(repr(t) for t in self.__templates),
        )

    def _template_match(self, value):
        return any(t.match(value) for t in self.__templates)


# #################### Part added from ts_coap plugin #################### #
class Not(Template):
    """
    Template class for Not
    """

    @typecheck
    def __init__(self, data: is_data):
        super().__init__(get_type(data))
        self.__data = store_data(data)

    @typecheck
    def _repr(self) -> str:
        return "%s(%s)" % (
            type(self).__name__,
            repr(self.__data),
        )

    @typecheck
    def _template_match(self, data: is_data) -> bool:
        return data not in self.__data



#!/usr/bin/env python3
#
#   (c) 2012  Universite de Rennes 1
#
# Contact address: <t3devkit@irisa.fr>
#
#
# This software is governed by the CeCILL license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import yaml

from ttproto.core.exceptions import ReaderError
from ttproto.core.typecheck import *


class YamlReader:
    """
    Reader class for yaml file or text
    """

    # Some class constants mainly for dumping format
    YAML_LINE_WIDTH = 70
    INDENTATION_SPACES_LENGTH = 4

    @typecheck
    def __init__(self, text: str, raw_text: bool = False):
        """
        Initialize yaml reader with the text can be a raw text or a filename

        :param text: The filename if file or a raw yaml text
        :param raw_text: True if raw text, default False for filename
        :type text: str
        :type raw_text: bool
        """
        self.__text = text
        self.__raw_text = raw_text
        self._yaml_as_dict = None
        self._flat_yaml = None

    @typecheck
    def __process_dict(self) -> dict:
        """
        Process the dictionary value of this yaml object

        :return: The python dict representation of this yaml object
        :rtype: dict
        """
        try:

            # If a raw text
            if self.__raw_text:
                return yaml.load(self.__text)

            # If a file
            else:
                with open(self.__text, 'r') as yaml_file:
                    return yaml.load(yaml_file)

        # If yaml error during decoding
        except yaml.YAMLError as ye:
            raise ReaderError(
                "YamlReader was unable to open and parse the %s file"
                %
                self.__text
            ) from ye

    @property
    def as_dict(self):
        """
        Singleton method to process the dictionary value of this yaml object

        :return: The python dict representation of this yaml object
        :rtype: dict
        """
        if not self._yaml_as_dict:
            self._yaml_as_dict = self.__process_dict()
        return self._yaml_as_dict

    @typecheck
    def __process_flat(self) -> str:
        """
        Process the flat value of this yaml object

        :return: The yaml flat representation
        :rtype: str
        """
        try:

            # If a raw text
            if self.__raw_text:
                to_dump = self.__text

            # If a file
            else:
                with open(self.__text, 'r') as yaml_file:
                    to_dump = yaml.load(yaml_file)

            # Return the loaded yaml file
            return yaml.dump(
                    to_dump,
                    width=self.YAML_LINE_WIDTH,
                    indent=self.INDENTATION_SPACES_LENGTH,
                    default_flow_style=False
                )

        # If yaml error during decoding
        except yaml.YAMLError as ye:
            raise ReaderError(
                "YamlReader was unable to flat the %s file"
                %
                self.__text
            ) from ye

    @property
    def as_flat_yaml(self):
        """
        Singleton method to process the flat value of this yaml object

        :return: The yaml flat representation
        :rtype: text
        """
        if not self._flat_yaml:
            self._flat_yaml = self.__process_flat()
        return self._flat_yaml



from ..common import *
import logging


class TD_COAP_CORE_21(CoAPTestCase):
    """
---
TD_COAP_CORE_21:
    cfg: CoAP_CFG_BASIC
    obj: Perform GET transaction containing the ETag option (CON mode)
    pre:
        - Server should offer a /validate resource which may be made to vary
            over time
        - Client & server supports ETag option
        - "The Client\u2019s cache must be purged"
    ref: '[COAP] 5.8.1, 5.10.7, 5.10.10, 12.1.12'
    seq:
        -   n: Verifying that client cache is empty
        -   s: "Client is requested to send a confirmable GET request to server\
                \u2019s resource"
        -   c:
            - 'The request sent request by the client contains:'
            -   - Type = 0 (CON)
                - Code = 1 (GET)
                - "Client-generated Message ID (\u2794 CMID)"
                - "Client-generated Token (\u2794 CTOK)"
                - Uri-Path option "validate"
                - No ETag option
        -   c:
            - 'Server sends response containing:'
            -   - Code = 2.05 (Content)
                - Message ID = CMID, Token = CTOK
                - "Option type = ETag, value = a value chosen by the server\
                  (\u2794 ETAG1)"
                - Non-empty Payload
        -   v: 'Client displays the response '
        -   n: Verifying client cache entry is still valid
        -   s: "Client is requested to send a confirmable GET request to server\
                \u2019s resource so as to check if the resource was updated"
        -   c:
            - 'The request sent by the client contains:'
            -   - Type = 0 (CON)
                - Code = 1 (GET)
                - "Another client-generated Message ID \u2260 CMID (\u2794 CMID2)"
                - "Client-generated Token which may or may not be \u2260 CTOK\
                   (\u2794 CTOK2)"
                - Uri-Path option "validate"
                - Option Type = ETag, value = ETAG1 (the ETag value received
                    in step 3)
        -   c:
            - 'Server sends response containing:'
            -   - Code = 2.03 (Valid)
                - Message ID = CMID2, Token = CTOK2
                - Option type = ETag, value = ETAG1
                - Empty Payload
        -   v: 'Client displays the response '
        -   n: Verifying that client cache entry is no longer valid
        -   s: "Update the content of the server\u2019s resource from a CoAP\
                client (either another client, or the testing client in a\
                separate transaction)"
        -   s: "Client is requested to send a confirmable GET request to server\
                \u2019s resource so as to check if the resource was updated"
        -   c:
            - 'The request sent by the client contains:'
            -   - Type = 0 (CON)
                - Code = 1 (GET)
                - "Another client-generated Message ID \u2260 CMID and \u2260\
                   CMID2 (\u2794 CMID3)"
                - "Client-generated Token which may or may not be \u2260 CTOK\
                   or CTOK2 (\u2794 CTOK3)"
                - Uri-Path option "validate"
                - Option Type = ETag, value = ETAG1 (the ETag value received
                    in step 3)
        -   c:
            - 'Server sends response containing:'
            -   - Code = 2.05 (Content)
                - Message ID = CMID3, Token = CTOK3
                - "Option type = ETag, value = another ETag value \u2260 ETAG1"
                - The payload of the requested resource, which should be different
                    from the payload in step 3
        -   v: Client displays the response
    """

    @classmethod
    @typecheck
    def get_stimulis(cls) -> list_of(Value):
        """
        Get the stimulis of this test case. This has to be be implemented into
        each test cases class.

        :return: The stimulis of this TC
        :rtype: [Value]

        .. note::
            Check the number/value of the uri query options or not?
        """
        return [
            CoAP(type='con', code='get', opt=Opt(CoAPOptionUriPath("validate"))),  # Step 1
            CoAP(type='con', code='get', opt=Opt(CoAPOptionUriPath("validate"))),  # Step 5
            CoAP(type='con', code='put', opt=Opt(CoAPOptionUriPath("validate"))),  # Step 9
            CoAP(type='con', code='get', opt=Opt(CoAPOptionUriPath("validate")))  # Step 10
        ]

    def run(self):
        # Part A
        # Step 2
        self.match("client", CoAP(type="con",
                                  code="get",
                                  opt=All(
                                      Opt(CoAPOptionUriPath("validate")),
                                      NoOpt(CoAPOptionETag()),
                                  )))
        CMID = self.coap["mid"]
        CTOK = self.coap["tok"]

        self.next_skip_ack()
        # Step 3
        if not self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                         code=2.05,
                                         mid=CMID,
                                         tok=CTOK,
                                         opt=Opt(CoAPOptionETag()),
                                         pl=Not(b""))):
            raise self.Stop()

        ETAG1 = self.coap["opt"][CoAPOptionETag]["val"]
        pl3 = self.coap["pl"]

        self.next_skip_ack(optional=True)

        # Part B
        # Step 6
        self.match("client", CoAP(type="con",
                                  code="get",
                                  opt=Opt(
                                      CoAPOptionUriPath("validate"),
                                      CoAPOptionETag(ETAG1),
                                  )))
        CMID2 = self.coap["mid"]
        CTOK2 = self.coap["tok"]

        if CMID2 != b"":
            if CMID2 == CMID:
                self.set_verdict("fail", "Message ID should be different")
        if CTOK2 != b"":
            if CTOK2 == CTOK:
                self.set_verdict("fail", "Token should be different")

        self.next_skip_ack()
        # Step 7
        self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                  code=2.03,
                                  mid=CMID2,
                                  tok=CTOK2,
                                  opt=Opt(CoAPOptionETag(ETAG1)),
                                  pl=b""))

        self.next_skip_ack(optional=True)

        # Part C

        if self.match("client", CoAP(code="put"), None):
            # allow an update from another client running on the same host
            self.next_skip_ack()
            self.match("server", CoAP(code=2.04))
            self.next_skip_ack(optional=True)
        # Step 11
        self.match("client", CoAP(type="con",
                                  code="get",
                                  opt=Opt(
                                      CoAPOptionUriPath("validate"),
                                      CoAPOptionETag(ETAG1),
                                  )))
        CMID3 = self.coap["mid"]
        CTOK3 = self.coap["tok"]

        if CMID3 != b"":
            if CMID3 == CMID or CMID3 == CMID2:
                self.set_verdict("fail", "Message ID should be different")
        if CTOK3 != b"":
            if CTOK3 == CTOK or CTOK3 == CTOK2:
                self.set_verdict("fail", "Token should be different")

        self.next_skip_ack()
        # Step 12
        self.match("server", CoAP(type=Any(CoAPType("con"), "ack"),
                                  code=2.05,
                                  mid=CMID3,
                                  tok=CTOK3,
                                  opt=Opt(CoAPOptionETag(Not(ETAG1))),
                                  pl=All(Not(b""), Not(pl3))), 'fail')

        self.next_skip_ack(optional=True)
